//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "PgpTestPgp.h"
#include "global.h"
#include <time.h>
HINSTANCE hSSLLib;
//---------------------------------------------------------------------------
// MISC PGP & CRYPTO
//---------------------------------------------------------------------------
byte PacketVersion(BUFFER *packet)
{
   int ptr = packet->ptr;
   buf_getc(packet); // ctb
   pgp_getsize(packet);
   byte ver=(byte)buf_getc(packet);
   packet->ptr = ptr;
   return ver;
}
//---------------------------------------------------------------------------
AnsiString TimeStamp(long unixtime)
{
   AnsiString tstamp = ctime(&unixtime);
   return tstamp.SubString(1,tstamp.Length()-1);
}
//---------------------------------------------------------------------------
int pgp_packettype(BUFFER *b, long *len)
{
  int ctb = buf_getc(b);
  switch (ctb >> 6)
  {
     case 2:
        /* old packet type */
        switch (ctb & 3)
        {
           case 0: *len = buf_getc(b); break;
           case 1: *len = buf_geti(b); break;
           case 2: *len = buf_getl(b); break;
           case 3: *len = b->length - b->ptr; break;
        }
        return (ctb >> 2) & 15;
     case 3:
        /* new packet type */
        *len = buf_getc(b);
        if (*len >= 192 && *len <= 223) *len = (*len - 192) * 256 + buf_getc(b) + 192; else
        if (*len == 255) *len = buf_getl(b); else
        if (*len > 223) return (-2); /* partial body length */
        return (ctb & 63);
  }
  return (-1);
}
//---------------------------------------------------------------------------
int pgp_getpacket(BUFFER *in, BUFFER *p)
/*  returns <0 = error, >0 = packet type */
{
  long len;
  int type = pgp_packettype(in, &len);
  if (type>0) buf_get(in, p, len);
  return (type);
}
//---------------------------------------------------------------------------
int pgp_packet(BUFFER *in, int type)
{
  int ctb;
  BUFFER *out;

  out = buf_new();
  ctb = 128 + (type << 2);
  if (in->length < 256 && type != PGP_PUBKEY && type != PGP_SECKEY &&
      type != PGP_SIG && type != PGP_PUBSUBKEY && type != PGP_SECSUBKEY
#ifdef MIMIC
      && type != PGP_ENCRYPTED
#endif
  )
  {
    buf_appendc(out, (char)ctb);
    buf_appendc(out, (char)in->length);
  }
#ifndef MIMIC
  else if (in->length < 65536)
#else
  else if ((type == PGP_PUBKEY || type == PGP_SECKEY || type == PGP_SIG
	   || type == PGP_SESKEY || type == PGP_PUBSUBKEY ||
	   type == PGP_SECSUBKEY) && in->length < 65536)
#endif
  {
    buf_appendc(out, (char)(ctb | 1));
    buf_appendi(out, in->length);
  } else {
    buf_appendc(out, (char)(ctb | 2));
    buf_appendl(out, in->length);
  }
  buf_cat(out, in);
  buf_move(in, out);
  buf_free(out);
  return (0);
}
//---------------------------------------------------------------------------
int pgpdb_getnext(BUFFER *keydb, BUFFER *key)
{
  long ptr;
  int type;
  BUFFER *p=buf_new();

  buf_clear(key);
  bool quitnow=false,havekey=false;
  if ((type=pgp_getpacket(keydb,p))>-1)
  {
     //int t=type;
     pgp_packet(p,type); buf_cat(key, p);
     while ((ptr=keydb->ptr,type=pgp_getpacket(keydb, p))>0)
     {
        havekey = true;
        switch (type)
        {
           case PGP_SECKEY:
           case PGP_PUBKEY:
              keydb->ptr = ptr;
              quitnow = true;
              break;
           case PGP_PUBSUBKEY:
           case PGP_SECSUBKEY:
	            break;
           case PGP_USERID:
              //sm(String(t)+"   "+String((char*)p->data));
              break;
        }
        if (quitnow) break;

        pgp_packet(p, type);
        buf_cat(key, p);
     }
  }
  buf_free(p);
  return (havekey ? 0 : -1);
}
//---------------------------------------------------------------------------
int mpi_get(BUFFER *b, BUFFER *mpi)
{
  buf_clear(mpi);
  int l=buf_geti(b);
  if (l <= 0 || b->ptr + (l + 7) / 8 > b->length) return (-1);
  buf_get(b, mpi, (l + 7) / 8);
  return (l);
}
//---------------------------------------------------------------------------
AnsiString BinaryByte(byte value, byte frombit, byte tobit)
{
   char buf[9];
   AnsiString bin=itoa(value,buf,2);
   bin = s("00000000").SubString(1,8-bin.Length())+bin;
   return bin.SubString(7-frombit+1,frombit-tobit+1);
}
//---------------------------------------------------------------------------
AnsiString PgpKeyIdText8(BUFFER *key, int offset)
{
   byte c;
   AnsiString text8;
   BUFFER *id = buf_new();
   PgpKeyId(key,offset,id);
   for (int i=0; i<4; i++)
   {
   	  c = id->data[4+i];
	  text8 += "0123456789ABCDEF" [ c >> 4  ];
	  text8 += "0123456789ABCDEF" [ c & 0xf ];
   }
   buf_free(id);
   return text8;
}
//---------------------------------------------------------------------------
int PgpKeyId(BUFFER *key, int offset, BUFFER *keyid)
{
  BUFFER *i,*k;
  int version,algo,j;

  i = buf_new();
  k = buf_new();
  key->ptr = offset;
  switch (version=buf_getc(key))
  {
     case 2:
     case 3:
        buf_getl(key); // timestamp
        buf_geti(key); // validity
        buf_getc(key); // algo
        mpi_get(key,i);
        break;
     case 4:
        buf_appendc(k, (char)version);
        buf_appendl(k, buf_getl(key));
        algo = buf_getc(key);
        buf_appendc(k, (char)algo);
        if (pgp_nummpi(algo) == 0) buf_rest(k, key); // works for public keys only
        else
           for (j = 0; j < pgp_nummpi(algo); j++)
           {
	          mpi_get(key,i);
 	          mpi_put(k,i);
           }
        buf_clear(i);
        buf_appendc(i, 0x99);
        buf_appendi(i, k->length);
        buf_cat(i, k);
        digest_sha1(i, i);
        break;
  }
  buf_clear(keyid);
  buf_append(keyid, i->data + i->length - 8, 8);
  buf_free(i);
  buf_free(k);
  return(0);
}
//---------------------------------------------------------------------------
AnsiString pgp_getkeyidasc(BUFFER *packet)
{
   byte c;
   AnsiString text8;
   BUFFER *id = buf_new();
   pgp_getkeyid(packet,id);
   for (int i=0; i<4; i++)
   {
   	  c = id->data[4+i];
	  text8 += "0123456789ABCDEF" [ c >> 4  ];
	  text8 += "0123456789ABCDEF" [ c & 0xf ];
   }
   buf_free(id);
   return text8;
}
//---------------------------------------------------------------------------
int pgp_getkeyid(BUFFER *packet, BUFFER *keyid)
{
  BUFFER *i,*k;
  int version,algo,j;
  buf_rewind(packet);
  byte ctb=(byte)buf_getc(packet);
  AnsiString bin=BinaryByte(ctb,5,2);
  pgp_getsize(packet);

  if (bin=="0110") // key certificate packet
  {
     i = buf_new();
     k = buf_new();
     switch (version=buf_getc(packet))
     {
        case 2:
        case 3:
           packet->ptr += ( 4 + 2 + 1 ); // timestamp, validity, algo
           mpi_get(packet,i);
           break;
        case 4:
           buf_appendc(k, (char)version);
           buf_appendl(k, buf_getl(packet));
           algo = buf_getc(packet);
           buf_appendc(k, (char)algo);
           if (pgp_nummpi(algo) == 0) buf_rest(k,packet); // works for public keys only
           else
              for (j = 0; j < pgp_nummpi(algo); j++)
              {
	             mpi_get(packet,i);
 	             mpi_put(k,i);
              }
           buf_clear(i);
           buf_appendc(i, 0x99);
           buf_appendi(i, k->length);
           buf_cat(i, k);
           digest_sha1(i, i);
           break;
     }
     buf_clear(keyid);
     buf_append(keyid, i->data + i->length - 8, 8);
     buf_free(i);
     buf_free(k);
     return 0;
  }
  
  if (bin=="0010") // signature packet
  {
     i = buf_new();
     k = buf_new();
     switch (version=buf_getc(packet))
     {
        case 2:
        case 3:
           packet->ptr += (1 + 1 + 4); // implicit, sigclass, timestamp
           buf_get(packet,i,8);
           break;
        case 4:
           packet->ptr += (1 + 1 + 1); // sigclass, hashalgo, pkaglo
           buf_get(packet,i,8);
           break;
     }
     buf_clear(keyid);
     buf_append(keyid, i->data + i->length - 8, 8);
     buf_free(i);
     buf_free(k);
     return 0;
  }
  return(0);
}
//---------------------------------------------------------------------------
void __fastcall SHA1(byte *b, int n, byte *m)
{
   _SHA1 sha1;
   if (!hSSLLib) return; // libeay32.dll
   if ((sha1 = (_SHA1)GetProcAddress(hSSLLib,"SHA1"))!=NULL)
      (sha1)(b,n,m);
}
//---------------------------------------------------------------------------
int mpi_put(BUFFER *buf, BUFFER *mpi)
{
  int l,i;
  l = mpi->length * 8;
  for (i = 7; i >= 0; i--)
    if (((mpi->data[0] >> i) & 1) == 1) {
      l -= 7 - i;
      break;
    }
  buf_appendi(buf, l);
  buf_cat(buf, mpi);
  return (l);
}
//---------------------------------------------------------------------------
int pgp_nummpi(int algo)
{
  switch (algo)
  {
     case PGP_ES_RSA: return (2);
     case PGP_S_DSA : return (4);
     case PGP_E_ELG : return (3);
  }
  return (0);
}
//---------------------------------------------------------------------------
int digestmem_sha1(byte *b, int n, BUFFER *md)
{
  byte m[SHA_DIGEST_LENGTH];

  SHA1(b, n, m);
  buf_reset(md);
  buf_append(md, m, SHA_DIGEST_LENGTH);
  return (0);
}
//---------------------------------------------------------------------------
int digest_sha1(BUFFER *b, BUFFER *md)
{
  return (digestmem_sha1(b->data, b->length, md));
}
//---------------------------------------------------------------------------
long lolf(BYTE CTB) // length-of-length field
{
   long result = 1;
   if ((CTB & 1) == 1) result = 2; else
   if ((CTB & 2) == 2) result = 4; else
   if ((CTB & 3) == 3) result = 0;
   return result;
}
//---------------------------------------------------------------------------
long getPacketSize(TFileStream *f, byte CTB)
{
   long l;
   BYTE tmp;
   BYTE *b = (BYTE*)&l;
   UINT *w = (UINT*)&l;
   long result = 0;

   f->Read(&l,lolf(CTB));
   switch (lolf(CTB)) {
      case 1: result = *b; break;
      case 2:
         tmp=b[0]; b[0]=b[1]; b[1]=tmp;
         result = *w;
         break;
      case 4: result = l; break;
   }
   return result;
}
//---------------------------------------------------------------------------
/*long pgp_getsize(BUFFER *packet)
{
   long l=0;
   buf_rewind(packet);
   byte ctb = buf_getc(packet);
   switch (lolf(ctb))
   {
      case 1: l = buf_getc(packet); break;
      case 2: l = buf_geti(packet); break;
      case 4: l = buf_getl(packet); break;
   }
   return l;
}*/
//---------------------------------------------------------------------------
long pgp_getsize(BUFFER *packet, byte ctb)
{
   long l=0;
   if (ctb==0)
   {
      buf_rewind(packet);
      ctb = (byte)buf_getc(packet);
   }
   switch (lolf(ctb))
   {
      case 1: l = buf_getc(packet); break;
      case 2: l = buf_geti(packet); break;
      case 4: l = buf_getl(packet); break;
   }
   return l;
}
//---------------------------------------------------------------------------
byte getCTB(TFileStream *f)
{
   BYTE ctb;
   f->Read(&ctb,1);           // 1-------
   return ((ctb & 128) == 128) ? ctb : BYTE(0);
}
//---------------------------------------------------------------------------
bool isSigCert(BYTE CTB)
{
   return ((CTB & 8) == 8);   // --0010--
}
//---------------------------------------------------------------------------
bool isKeyCert(BYTE CTB)
{
   return ((CTB & 24) == 24); // --0110--
}
//---------------------------------------------------------------------------
bool isUserId(BYTE CTB)
{
   return ((CTB & 52) == 52); // --1101--
}
//---------------------------------------------------------------------------
bool isTrustByte(BYTE CTB)
{
   return ((CTB & 48) == 48); // --1100--
}
//---------------------------------------------------------------------------


